@@ -1,7 +1,7 @@
-/* @(#)readcd.c	1.14 00/04/24 Copyright 1987 J. Schilling */
+/* @(#)readcd.c	1.20 00/07/20 Copyright 1987 J. Schilling */
 #ifndef lint
 static	char sccsid[] =
-	"@(#)readcd.c	1.14 00/04/24 Copyright 1987 J. Schilling";
+	"@(#)readcd.c	1.20 00/07/20 Copyright 1987 J. Schilling";
 #endif
 /*
  *	Skeleton for the use of the scg genearal SCSI - driver
@@ -32,6 +32,7 @@ static	char sccsid[] =
 #include <strdefs.h>
 #include <fctldefs.h>
 #include <timedefs.h>
+#include <schily.h>
 
 #include <scg/scgcmd.h>
 #include <scg/scsireg.h>
@@ -39,7 +40,19 @@ static	char sccsid[] =
 
 #include "cdrecord.h"
 
-char	cdr_version[] = "1.8.1";
+char	cdr_version[] = "1.9";
+
+#if     defined(PROTOTYPES)
+#define UINT_C(a)	(a##u)
+#define ULONG_C(a)	(a##ul)
+#define USHORT_C(a)	(a##uh)
+#define CONCAT(a,b)	a##b
+#else
+#define UINT_C(a)	((unsigned)(a))
+#define ULONG_C(a)	((unsigned long)(a))
+#define USHORT_C(a)	((unsigned short)(a))
+#define CONCAT(a,b)	a/**/b
+#endif
 
 extern	BOOL	getlong		__PR((char *, long *, long, long));
 extern	BOOL	getint		__PR((char *, int *, int, int));
@@ -47,17 +60,38 @@ extern	BOOL	getint		__PR((char *, int *, int, int));
 typedef struct {
 	long	start;
 	long	end;
+	long	sptr;		/* sectors per transfer */
+	BOOL	askrange;
 	char	*name;
 } parm_t;
 
-EXPORT	BOOL cvt_cyls		__PR((void));
-EXPORT	BOOL cvt_bcyls		__PR((void));
-EXPORT	void print_defect_list	__PR((void));
-LOCAL	void usage		__PR((int ret));
-EXPORT	int main		__PR((int ac, char **av));
-LOCAL	int prstats		__PR((void));
-LOCAL	void dorw		__PR((SCSI *scgp, char *filename, char* sectors));
-LOCAL	void doit		__PR((SCSI *scgp));
+typedef struct {
+	int	errors;
+	int	c2_errors;
+	int	c2_errsecs;
+	int	secsize;
+	BOOL	ismmc;
+} rparm_t;
+
+EXPORT	BOOL	cvt_cyls	__PR((void));
+EXPORT	BOOL	cvt_bcyls	__PR((void));
+EXPORT	void	print_defect_list __PR((void));
+LOCAL	void	usage		__PR((int ret));
+EXPORT	int	main		__PR((int ac, char **av));
+LOCAL	int	prstats		__PR((void));
+LOCAL	void	dorw		__PR((SCSI *scgp, char *filename, char* sectors));
+LOCAL	void	doit		__PR((SCSI *scgp));
+LOCAL	void	read_disk	__PR((SCSI *scgp, parm_t *parmp));
+LOCAL	void	readc2_disk	__PR((SCSI *scgp, parm_t *parmp));
+
+LOCAL	int	fread_data	__PR((SCSI *scgp, rparm_t *rp, caddr_t bp, long addr, int cnt));
+LOCAL	int	bits		__PR((int c));
+LOCAL	int	bitidx		__PR((int c));
+LOCAL	int	fread_c2	__PR((SCSI *scgp, rparm_t *rp, caddr_t bp, long addr, int cnt));
+
+LOCAL	int	fdata_null	__PR((rparm_t *rp, caddr_t bp, long addr, int cnt));
+LOCAL	int	fdata_c2	__PR((rparm_t *rp, caddr_t bp, long addr, int cnt));
+
 #ifdef	used
 LOCAL	int read_scsi_g1	__PR((SCSI *scgp, caddr_t bp, long addr, int cnt));
 #endif
@@ -69,16 +103,22 @@ EXPORT	int	write_g1	__PR((SCSI *scgp, caddr_t bp, long addr, int cnt));
 #ifdef	used
 LOCAL	void	Xrequest_sense	__PR((SCSI *scgp));
 #endif
-LOCAL	int	rrr		__PR((SCSI *scgp, caddr_t bp, long addr, long cnt));
-LOCAL	void	read_disk	__PR((SCSI	*scgp, parm_t *parmp));
+LOCAL	int	read_retry	__PR((SCSI *scgp, caddr_t bp, long addr, long cnt,
+					int (*rfunc)(SCSI *scgp, rparm_t *rp, caddr_t bp, long addr, int cnt),
+					rparm_t *rp
+					));
+LOCAL	void	read_generic	__PR((SCSI *scgp, parm_t *parmp,
+					int (*rfunc)(SCSI *scgp, rparm_t *rp, caddr_t bp, long addr, int cnt),
+					rparm_t *rp,
+					int (*dfunc)(rparm_t *rp, caddr_t bp, long addr, int cnt)
+));
 LOCAL	void	write_disk	__PR((SCSI *scgp, parm_t *parmp));
 LOCAL	int	choice		__PR((int n));
 LOCAL	void	ra		__PR((SCSI *scgp));
 
-#ifndef	TRUE
-#	define	TRUE	1
-#	define	FALSE	0
-#endif
+EXPORT	int	read_da		__PR((SCSI *scgp, caddr_t bp, long addr, int cnt, int framesize, int subcode));
+EXPORT	int	read_cd		__PR((SCSI *scgp, caddr_t bp, long addr, int cnt, int framesize, int data, int subch));
+
 
 struct timeval	starttime;
 struct timeval	stoptime;
@@ -86,10 +126,17 @@ struct timeval	stoptime;
 char	*Sbuf;
 long	Sbufsize;
 
+/*#define	MAX_RETRY	32*/
+#define	MAX_RETRY	128
+
 int	help;
 BOOL	is_suid;
 BOOL	is_cdrom;
 BOOL	do_write;
+BOOL	c2scan;
+BOOL	fulltoc;
+BOOL	noerror;
+int	retries = MAX_RETRY;
 
 struct	scsi_format_data fmt;
 
@@ -108,14 +155,17 @@ usage(ret)
 	error("\tf=filename	Name of file to read/write\n");
 	error("\tsectors=range	Range of sectors to read/write\n");
 	error("\t-w		Switch to write mode\n");
+	error("\t-c2scan		Do a C2 error scan\n");
 	error("\tkdebug=#,kd=#\tdo Kernel debugging\n");
 	error("\t-v		increment general verbose level by one\n");
 	error("\t-V		increment SCSI command transport verbose level by one\n");
 	error("\t-silent,-s\tdo not print status of erreneous commands\n");
+	error("\t-noerror	do not abort on error\n");
+	error("\tretries=#	set retry count (default is %d)\n", retries);
 	exit(ret);
 }	
 
-char	opts[]   = "kdebug#,kd#,verbose+,v+,Verbose,V+,silent,s,debug,help,h,version,dev*,f*,sectors*,w";
+char	opts[]   = "kdebug#,kd#,verbose+,v+,Verbose,V+,silent,s,debug,help,h,version,dev*,sectors*,w,c2scan,fulltoc,noerror,retries#,f*";
 
 EXPORT int
 main(ac, av)
@@ -150,7 +200,9 @@ main(ac, av)
 			&lverbose, &lverbose,
 			&silent, &silent, &debug,
 			&help, &help, &pversion,
-			&dev, &filename, &sectors, &do_write) < 0) {
+			&dev, &sectors, &do_write,
+			&c2scan, &fulltoc,
+			&noerror, &retries, &filename) < 0) {
 		errmsgno(EX_BAD, "Bad flag: %s.\n", cav[0]);
 		usage(EX_BAD);
 	}
@@ -208,7 +260,6 @@ main(ac, av)
 
 		if ((scgp = open_scsi(dev, errstr, sizeof(errstr), debug, lverbose)) == (SCSI *)0)
 			comerr("%s%sCannot open SCSI driver.\n", errstr, errstr[0]?". ":"");
-/*		scsi_settimeout(scgp, timeout);*/
 	} else {
 		if (scsibus == -1 && target >= 0 && lun >= 0)
 			scsibus = 0;
@@ -231,6 +282,8 @@ main(ac, av)
 	scgp->verbose = verbose;
 	scgp->debug = debug;
 	scgp->kdebug = kdebug;
+/*	scsi_settimeout(scgp, timeout);*/
+	scsi_settimeout(scgp, 40);
 
 	Sbufsize = scsi_bufsize(scgp, 256*1024L);
 	if ((Sbuf = scsi_getbuf(scgp, Sbufsize)) == NULL)
@@ -259,7 +312,7 @@ main(ac, av)
 			comerrno(EX_BAD, "Not root. Will only work on CD-ROM in suid mode\n");
 	}
 
-	if (filename || sectors) {
+	if (filename || sectors || c2scan || fulltoc) {
 		dorw(scgp, filename, sectors);
 	} else {
 		doit(scgp);
@@ -267,6 +320,9 @@ main(ac, av)
 	return (0);
 }
 
+/*
+ * Return milliseconds since start time.
+ */
 LOCAL int
 prstats()
 {
@@ -289,7 +345,6 @@ prstats()
 	}
 
 	error("Time total: %d.%03dsec\n", sec, usec/1000);
-/*	return (sec + (usec / 500000));*/
 	return (1000*sec + (usec / 1000));
 }
 
@@ -304,6 +359,8 @@ dorw(scgp, filename, sectors)
 
 	params.start = 0;
 	params.end = -1;
+	params.sptr = -1;
+	params.askrange = FALSE;
 	params.name = NULL;
 
 	if (filename)
@@ -318,7 +375,11 @@ dorw(scgp, filename, sectors)
 	if (!wait_unit_ready(scgp, 60))
 		comerr("Device not ready.\n");
 
-	if (do_write)
+	if (c2scan) {
+		if (params.name == NULL)
+			params.name = "/dev/null";
+		readc2_disk(scgp, &params);
+	} else if (do_write)
 		write_disk(scgp, &params);
 	else
 		read_disk(scgp, &params);
@@ -329,6 +390,13 @@ doit(scgp)
 	SCSI	*scgp;
 {
 	int	i = 0;
+	parm_t	params;
+
+	params.start = 0;
+	params.end = -1;
+	params.sptr = -1;
+	params.askrange = TRUE;
+	params.name = "/dev/null";
 
 	for(;;) {
 		if (!wait_unit_ready(scgp, 60))
@@ -336,18 +404,179 @@ doit(scgp)
 
 		printf("0:read 1:veri   2:erase   3:read buffer 4:cache 5:ovtime 6:cap\n");
 		printf("7:wne  8:floppy 9:verify 10:checkcmds  11:read disk 12:write disk\n");
-		printf("13:scsireset 14:seektest 15: readda\n");
+		printf("13:scsireset 14:seektest 15: readda 16: reada 17: c2err\n");
 
-		getint("Enter selection:", &i, 0, 15);
+		getint("Enter selection:", &i, 0, 20);
 		switch (i) {
 
 		case 11:	read_disk(scgp, 0);	break;
 		case 12:	write_disk(scgp, 0);	break;
 		case 15:	ra(scgp);		break;
+/*		case 16:	reada_disk(scgp, 0, 0);	break;*/
+		case 17:	readc2_disk(scgp, &params);	break;
 		}
 	}
 }
 
+LOCAL void
+read_disk(scgp, parmp)
+	SCSI	*scgp;
+	parm_t	*parmp;
+{
+	rparm_t	rp;
+
+	read_capacity(scgp);
+	rp.errors = 0;
+	rp.c2_errors = 0;
+	rp.c2_errsecs = 0;
+	rp.secsize = scgp->cap->c_bsize;
+
+	read_generic(scgp, parmp, fread_data, &rp, fdata_null);
+}
+
+
+char	zeroblk[512];
+
+LOCAL void
+readc2_disk(scgp, parmp)
+	SCSI	*scgp;
+	parm_t	*parmp;
+{
+	rparm_t	rp;
+
+	rp.errors = 0;
+	rp.c2_errors = 0;
+	rp.c2_errsecs = 0;
+	rp.secsize = 2352 + 294;
+	rp.ismmc = is_mmc(scgp, NULL);
+
+
+	read_generic(scgp, parmp, fread_c2, &rp, fdata_c2);
+
+	printf("C2 errors total: %d bytes in %d sectors on disk\n", rp.c2_errors, rp.c2_errsecs);
+	printf("C2 errors rate: %f%% \n", (100.0*rp.c2_errors)/scgp->cap->c_baddr/2352);
+}
+
+LOCAL int
+fread_data(scgp, rp, bp, addr, cnt)
+	SCSI	*scgp;
+	rparm_t	*rp;
+	caddr_t	bp;
+	long	addr;
+	int	cnt;
+{
+	return (read_g1(scgp, bp, addr, cnt));
+}
+
+
+int bits(c)
+	int	c;
+{
+	int	n = 0;
+
+	if (c & 0x01)
+		n++;
+	if (c & 0x02)
+		n++;
+	if (c & 0x04)
+		n++;
+	if (c & 0x08)
+		n++;
+	if (c & 0x10)
+		n++;
+	if (c & 0x20)
+		n++;
+	if (c & 0x40)
+		n++;
+	if (c & 0x80)
+		n++;
+	return (n);
+}
+
+int bitidx(c)
+	int	c;
+{
+	if (c & 0x80)
+		return (0);
+	if (c & 0x40)
+		return (1);
+	if (c & 0x20)
+		return (2);
+	if (c & 0x10)
+		return (3);
+	if (c & 0x08)
+		return (4);
+	if (c & 0x04)
+		return (5);
+	if (c & 0x02)
+		return (6);
+	if (c & 0x01)
+		return (7);
+	return (-1);
+}
+
+LOCAL int
+fread_c2(scgp, rp, bp, addr, cnt)
+	SCSI	*scgp;
+	rparm_t	*rp;
+	caddr_t	bp;
+	long	addr;
+	int	cnt;
+{
+	if (rp->ismmc) {
+		return (read_cd(scgp, bp, addr, cnt, rp->secsize,
+			/* Sync + all headers + user data + EDC/ECC + C2 */
+/*			(1 << 7 | 3 << 5 | 1 << 4 | 1 << 3 | 2 << 1),*/
+			(1 << 7 | 3 << 5 | 1 << 4 | 1 << 3 | 1 << 1),
+        		/* without subchannels */
+			0));
+	} else {
+		return (read_da(scgp, bp, addr, cnt, rp->secsize,
+			/* Sync + all headers + user data + EDC/ECC + C2 */
+			0x04));
+	}
+}
+
+LOCAL int
+fdata_null(rp, bp, addr, cnt)
+	rparm_t	*rp;
+	caddr_t	bp;
+	long	addr;
+	int	cnt;
+{
+	return (0);
+}
+
+LOCAL int
+fdata_c2(rp, bp, addr, cnt)
+	rparm_t	*rp;
+	caddr_t	bp;
+	long	addr;
+	int	cnt;
+{
+	int	i;
+	int	j;
+	int	k;
+	char	*p;
+
+	p = &bp[2352];
+
+	for (i=0; i < cnt; i++, p += (2352+294)) {
+/*		scsiprbytes("XXX ", p, 294);*/
+		if ((j = cmpbytes(p, zeroblk, 294)) < 294) {
+			printf("C2 in sector: %3ld first at byte: %4d (0x%02X)", addr+i,
+				j*8 + bitidx(p[j]), p[j]&0xFF);
+			for (j=0,k=0; j < 294; j++)
+				k += bits(p[j]);
+			printf(" total: %4d errors\n", k);
+/*			scsiprbytes("XXX ", p, 294);*/
+			rp->c2_errors += k;
+			rp->c2_errsecs++;
+		}
+	}	
+	return (0);
+}
+
 #ifdef	used
 LOCAL int
 read_scsi_g1(scgp, bp, addr, cnt)
@@ -493,20 +722,24 @@ Xrequest_sense(scgp)
 }
 #endif
 
-/*#define	MAX_RETRY	32*/
-#define	MAX_RETRY	128
-
 LOCAL int
-rrr(scgp, bp, addr, cnt)
+read_retry(scgp, bp, addr, cnt, rfunc, rp)
 	SCSI	*scgp;
 	caddr_t	bp;
 	long	addr;
 	long	cnt;
+	int	(*rfunc)__PR((SCSI *scgp, rparm_t *rp, caddr_t bp, long addr, int cnt));
+	rparm_t	*rp;
 {
 	int	secsize = scgp->cap->c_bsize;
 	int	try = 0;
 	int	err;
-	char	dummybuf[4000];
+	char	dummybuf[8192];
+
+	if (secsize > sizeof(dummybuf)) {
+		errmsgno(EX_BAD, "Cannot retry, sector size %d too big.\n", secsize);
+		return (-1);
+	}
 
 	errmsgno(EX_BAD, "Retrying from sector %d.\n", addr);
 	while (cnt > 0) {
@@ -517,23 +750,25 @@ rrr(scgp, bp, addr, cnt)
 				if ((try % 8) == 0) {
 					error("+");
 					scgp->silent++;
-					read_g1(scgp, dummybuf, scgp->cap->c_baddr, 1);
+					(*rfunc)(scgp, rp, dummybuf, scgp->cap->c_baddr, 1);
 					scgp->silent--;
 				} else if ((try % 4) == 0) {
 					error("-");
 					scgp->silent++;
-					read_g1(scgp, dummybuf, 0, 1);
+					(*rfunc)(scgp, rp, dummybuf, 0, 1);
 					scgp->silent--;
 				} else {
 					error("~");
 					scgp->silent++;
-					read_g1(scgp, dummybuf, choice(scgp->cap->c_baddr), 1);
+					(*rfunc)(scgp, rp, dummybuf, choice(scgp->cap->c_baddr), 1);
 					scgp->silent--;
 				}
 			}
 
+			fillbytes(bp, secsize, 0);
+
 			scgp->silent++;
-			err = read_g1(scgp, bp, addr, 1);
+			err = (*rfunc)(scgp, rp, bp, addr, 1);
 			scgp->silent--;
 
 			if (err < 0) {
@@ -543,16 +778,17 @@ rrr(scgp, bp, addr, cnt)
 			} else {
 				break;
 			}
-		} while (++try < MAX_RETRY);
+		} while (++try < retries);
 
-		if (try >= MAX_RETRY) {
+		if (try >= retries) {
 			error("\n");
-			errmsgno(err, "Error not corrected.\n");
-			return (-1);
+			errmsgno(err, "Error on sector %d not corrected. Total of %d errors.\n", addr, ++rp->errors);
+			if (!noerror)
+				return (-1);
 		} else {
 			if (try > 1) {
 				error("\n");
-				errmsgno(EX_BAD, "Error corrected after %d tries.\n", try);
+				errmsgno(EX_BAD, "Error corrected after %d tries. Total of %d errors.\n", try, rp->errors);
 			}
 		}
 		try = 0;
@@ -564,65 +800,89 @@ rrr(scgp, bp, addr, cnt)
 }
 
 LOCAL void
-read_disk(scgp, parmp)
+read_generic(scgp, parmp, rfunc, rp, dfunc)
 	SCSI	*scgp;
 	parm_t	*parmp;
+	int	(*rfunc)__PR((SCSI *scgp, rparm_t *rp, caddr_t bp, long addr, int cnt));
+	rparm_t	*rp;
+	int	(*dfunc)__PR((rparm_t *rp, caddr_t bp, long addr, int cnt));
 {
 	char	filename[512];
-	char	*defname = "disk.out";
+	char	*defname = NULL;
 	FILE	*f;
 	long	addr;
-	long	cnt;
-	long	end;
-	int	nsec;
+	long	num;
+	long	end = 0L;
 	long	start = 0L;
+	long	cnt;
+	int	msec;
 	int	err = 0;
+	BOOL	askrange = FALSE;
+	BOOL	isrange = FALSE;
+	int	secsize = rp->secsize;
 
 	if (is_suid) {
 		if (scgp->inq->type != INQ_ROMD)
 			comerrno(EX_BAD, "Not root. Will only read from CD in suid mode\n");
 	}
 
-	filename[0] ='\0';
-	read_capacity(scgp);
-	end = scgp->cap->c_baddr + 1;
+	if (parmp == NULL || parmp->askrange)
+		askrange = TRUE;
+	if (parmp != NULL && !askrange && (parmp->start <= parmp->end))
+		isrange = TRUE;
 
-	if (end <= 0)
+	filename[0] ='\0';
+	if (read_capacity(scgp) >= 0)
+		end = scgp->cap->c_baddr + 1;
+
+	if (end <= 0 || isrange || (askrange && scsi_yes("Ignore disk size? ")))
 		end = 10000000;	/* Hack to read empty (e.g. blank=fast) disks */
 
 	if (parmp) {
 		if (parmp->name)
 			defname = parmp->name;
-		error("Copy from SCSI (%d,%d,%d) disk to file '%s'\n",
+		if (defname != NULL) {
+			error("Copy from SCSI (%d,%d,%d) disk to file '%s'\n",
 					scgp->scsibus, scgp->target, scgp->lun,
 					defname);
+		}
 
 		addr = start = parmp->start;
 		if (parmp->end != -1 && parmp->end < end)
 			end = parmp->end;
-		cnt = Sbufsize / scgp->cap->c_bsize;
+		cnt = Sbufsize / secsize;
 	}
-else {
 
-	error("Copy from SCSI (%d,%d,%d) disk to file\n",
+	if (defname == NULL) {
+		defname = "disk.out";
+		error("Copy from SCSI (%d,%d,%d) disk to file\n",
 					scgp->scsibus, scgp->target, scgp->lun);
-	error("Enter filename [%s]: ", defname);flush();
-	(void)getline(filename, sizeof(filename));
+		error("Enter filename [%s]: ", defname);flush();
+		(void)getline(filename, sizeof(filename));
+	}
 
-	addr = start;
-	getlong("Enter starting sector for copy:", &addr, start, end-1);
-	start = addr;
-	cnt = end - addr;
-	getlong("Enter number of sectors to copy:", &cnt, 1L, cnt);
-	end = addr + cnt;
+	if (askrange) {
+		addr = start;
+		getlong("Enter starting sector for copy:", &addr, start, end-1);
+/*		getlong("Enter starting sector for copy:", &addr, -300, end-1);*/
+		start = addr;
+	}
 
-	cnt = Sbufsize / scgp->cap->c_bsize;
-	getlong("Enter number of sectors per copy:", &cnt, 1L, cnt);
-/*	error("end:  %8ld\n", end);*/
-}
+	if (askrange) {
+		num = end - addr;
+		getlong("Enter number of sectors to copy:", &num, 1L, num);
+		end = addr + num;
+	}
+
+	if (askrange) {
+/* XXX askcnt */
+		cnt = Sbufsize / secsize;
+		getlong("Enter number of sectors per copy:", &cnt, 1L, cnt);
+	}
 
 	if (filename[0] == '\0')
-		strcpy(filename, defname);
+		strncpy(filename, defname, sizeof(filename));
+	filename[sizeof(filename)-1] = '\0';
 	if (streql(filename, "-")) {
 		f = stdout;
 #if	defined(__CYGWIN32__) || defined(__EMX__)
@@ -642,15 +902,16 @@ else {
 
 		error("addr: %8d cnt: %d\r", addr, cnt);
 
-/*		if (read_scsi(scgp, Sbuf, addr, cnt) < 0) {*/
-		if (read_g1(scgp, Sbuf, addr, cnt) < 0) {
+		if ((*rfunc)(scgp, rp, Sbuf, addr, cnt) < 0) {
 			err = scgp->scmd->ux_errno;
 			error("\n");
 			errmsgno(err, "Cannot read source disk\n");
-			if (rrr(scgp, Sbuf, addr, cnt) < 0)
+
+			if (read_retry(scgp, Sbuf, addr, cnt, rfunc, rp) < 0)
 				goto out;
 		}
-		if (filewrite(f, Sbuf, cnt * scgp->cap->c_bsize) < 0) {
+		(*dfunc)(rp, Sbuf, addr, cnt);
+		if (filewrite(f, Sbuf, cnt * secsize) < 0) {
 			err = geterrno();
 			error("\n");
 			errmsgno(err, "Cannot write '%s'\n", filename);
@@ -660,13 +921,16 @@ else {
 	error("addr: %8d", addr);
 out:
 	error("\n");
-	nsec = prstats();
-/*	error("%d\n",nsec);*/
+	msec = prstats();
+#ifdef	OOO
 	error("Read %.2f kB at %.1f kB/sec.\n",
-		(addr - start)/(1024.0/scgp->cap->c_bsize),
-		((addr - start)/(1024.0/scgp->cap->c_bsize)) / (0.001*nsec));
-	if (err)
-		exit(err);
+		(double)(addr - start)/(1024.0/scgp->cap->c_bsize),
+		(double)((addr - start)/(1024.0/scgp->cap->c_bsize)) / (0.001*msec));
+#else
+	error("Read %.2f kB at %.1f kB/sec.\n",
+		(double)(addr - start)/(1024.0/secsize),
+		(double)((addr - start)/(1024.0/secsize)) / (0.001*msec));
+#endif
 }
 
 LOCAL void
@@ -681,12 +945,13 @@ write_disk(scgp, parmp)
 	long	cnt;
 	long	amt;
 	long	end;
-	int	nsec;
+	int	msec;
 	int	start;
 
 	if (is_suid)
 		comerrno(EX_BAD, "Not root. Will not write in suid mode\n");
 
+	filename[0] ='\0';
 	read_capacity(scgp);
 	end = scgp->cap->c_baddr + 1;
 
@@ -696,7 +961,7 @@ write_disk(scgp, parmp)
 	if (parmp) {
 		if (parmp->name)
 			defname = parmp->name;
-		error("Copy from file '%s' tp SCSI (%d,%d,%d) disk\n",
+		error("Copy from file '%s' to SCSI (%d,%d,%d) disk\n",
 					defname,
 					scgp->scsibus, scgp->target, scgp->lun);
 
@@ -725,7 +990,8 @@ else {
 }
 
 	if (filename[0] == '\0')
-		strcpy(filename, defname);
+		strncpy(filename, defname, sizeof(filename));
+	filename[sizeof(filename)-1] = '\0';
 	if (streql(filename, "-")) {
 		f = stdin;
 #if	defined(__CYGWIN32__) || defined(__EMX__)
@@ -756,10 +1022,10 @@ else {
 					"Cannot write destination disk\n");
 	}
 	error("addr: %8d\n", addr);
-	nsec = prstats();
+	msec = prstats();
 	error("Wrote %.2f kB at %.1f kB/sec.\n",
 		(double)(addr - start)/(1024.0/scgp->cap->c_bsize),
-		(double)((addr - start)/(1024.0/scgp->cap->c_bsize)) / nsec);
+		(double)((addr - start)/(1024.0/scgp->cap->c_bsize)) / (0.001*msec));
 }
 
 LOCAL int
@@ -790,7 +1056,7 @@ ra(scgp)
 /*	long	addr = 0L;*/
 /*	long	cnt;*/
 /*	long	end;*/
-/*	int	nsec;*/
+/*	int	msec;*/
 /*	int	start;*/
 /*	int	err = 0;*/
 
@@ -805,13 +1071,18 @@ ra(scgp)
 	fclose(f);
 }
 
-#ifdef	XXX
+#define	g5x_cdblen(cdb, len)	((cdb)->count[0] = ((len) >> 16L)& 0xFF,\
+				 (cdb)->count[1] = ((len) >> 8L) & 0xFF,\
+				 (cdb)->count[2] = (len) & 0xFF)
+
 EXPORT int
-read_da(scgp, bp, addr, cnt)
+read_da(scgp, bp, addr, cnt, framesize, subcode)
 	SCSI	*scgp;
 	caddr_t	bp;
 	long	addr;
 	int	cnt;
+	int	framesize;
+	int	subcode;
 {
 	register struct	scg_cmd	*scmd = scgp->scmd;
 
@@ -820,19 +1091,53 @@ read_da(scgp, bp, addr, cnt)
 
 	fillbytes((caddr_t)scmd, sizeof(*scmd), '\0');
 	scmd->addr = bp;
-	scmd->size = cnt*scgp->cap->c_bsize;
+	scmd->size = cnt*framesize;
 	scmd->flags = SCG_RECV_DATA|SCG_DISRE_ENA;
 	scmd->cdb_len = SC_G5_CDBLEN;
 	scmd->sense_len = CCS_SENSE_LEN;
 	scmd->target = scgp->target;
-/*	scmd->cdb.g5_cdb.cmd = SC_EREAD;*/
-	scmd->cdb.g5_cdb.cmd = 0xd9;
+	scmd->cdb.g5_cdb.cmd = 0xd8;
 	scmd->cdb.g5_cdb.lun = scgp->lun;
 	g5_cdbaddr(&scmd->cdb.g1_cdb, addr);
 	g5_cdblen(&scmd->cdb.g1_cdb, cnt);/* XXX subscript out of range ?? */
+	g5x_cdblen(&scmd->cdb.g1_cdb, cnt);/* XXX subscript out of range ?? */
+	scmd->cdb.g5_cdb.res10 = subcode;
 	
-	scgp->cmdname = "read_g1";
+	scgp->cmdname = "read_da";
 
 	return (scsicmd(scgp));
 }
-#endif
+
+EXPORT int
+read_cd(scgp, bp, addr, cnt, framesize, data, subch)
+	SCSI	*scgp;
+	caddr_t	bp;
+	long	addr;
+	int	cnt;
+	int	framesize;
+	int	data;
+	int	subch;
+{
+	register struct	scg_cmd	*scmd = scgp->scmd;
+
+	fillbytes((caddr_t)scmd, sizeof(*scmd), '\0');
+	scmd->addr = bp;
+	scmd->size = cnt*framesize;
+	scmd->flags = SCG_RECV_DATA|SCG_DISRE_ENA;
+	scmd->cdb_len = SC_G5_CDBLEN;
+	scmd->sense_len = CCS_SENSE_LEN;
+	scmd->target = scgp->target;
+	scmd->cdb.g5_cdb.cmd = 0xBE;
+	scmd->cdb.g5_cdb.lun = scgp->lun;
+        scmd->cdb.g5_cdb.res = 0;	/* expected sector type field ALL */
+	g5_cdbaddr(&scmd->cdb.g5_cdb, addr);
+	g5x_cdblen(&scmd->cdb.g5_cdb, cnt);
+
+	scmd->cdb.g5_cdb.count[3] = data & 0xFF;
+        scmd->cdb.g5_cdb.res10 = subch & 0x07;
+
+	scgp->cmdname = "read_cd";
+
+	return (scsicmd(scgp));
+}
+
